/*

  Copyright (C) 2002, 2005, 2006  Dmitry V. Levin <ldv@altlinux.org>

  The Window Manager's Selector.

  This file is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
*/

#ifndef _GNU_SOURCE
# define _GNU_SOURCE	1
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <error.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <dirent.h>
#include <locale.h>
#include <libintl.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#define PACKAGE "wm-select"
#define LOCALEDIR "/usr/share/locale"
#define _(arg) gettext (arg)

#define WM_DIR "/etc/X11/wmsession.d"

#ifdef WITH_GTK2
  #define FONT_TYPE PangoFontDescription
  #define FONT_FUNC pango_font_description_from_string
  #define FONT_ATTR font_desc
  #define WM_SELECT_LABEL_FONT "Serif 14"
  #define WM_SELECT_WELCOME_FONT "Serif 24"
#else /* ! WITH_GTK2 */
  #define FONT_TYPE GdkFont
  #define FONT_FUNC gdk_font_load
  #define FONT_ATTR font
  #define WM_SELECT_LABEL_FONT "-*-times-medium-r-*--*-140-*-*-*-*-*-*"
  #define WM_SELECT_WELCOME_FONT "-*-times-bold-r-*--*-240-*-*-*-*-*-*"
#endif /* WITH_GTK2 */

static const char *selected;
static const char *preferred;
static int save_prefs;

static char *
getval(char *line, const char *name)
{
	size_t  namelen = strlen(name);

	if (!strncmp(line, name, namelen))
	{
		char   *val = line + namelen;
		size_t  len = strlen(val);

		if (len && ('\n' == val[len - 1]))
			val[--len] = '\0';
		return val;
	}

	return 0;
}

static char *
xasprintf(char **ptr, const char *fmt, ...)
{
	va_list arg;

	va_start(arg, fmt);
	if (vasprintf(ptr, fmt, arg) < 0)
		error(EXIT_FAILURE, errno, "vasprintf");
	va_end(arg);

	return *ptr;
}

static char *
GetRCF()
{
	const char *home = getenv("HOME");
	char   *fname = 0;

	if (!home || !*home)
		return 0;

	xasprintf(&fname, "%s/.%s", home, program_invocation_short_name);

	return fname;
}

static void
SavePrefs()
{
	char   *fname;
	FILE   *fp;

	if (!save_prefs)
		return;

	fname = GetRCF();
	if (!fname)
		return;

	fp = fopen(fname, "w");
	free(fname);
	if (!fp)
		return;

	fprintf(fp, "EXEC=%s\n", selected);
	fclose(fp);
}

static void
ExecSelected()
{
	if (selected)
	{
		SavePrefs();
		// error (EXIT_SUCCESS, 0, "executing: %s", selected);
		execl(selected, basename(selected), NULL);
		error(EXIT_SUCCESS, errno, "execl: %s", selected);
		selected = 0;
	}
}

static void
ReadProgramConfig()
{
	char   *fname = GetRCF();
	FILE   *fp;
	char    buf[BUFSIZ];

	if (!fname)
		return;

	fp = fopen(fname, "r");
	free(fname);
	if (!fp)
		return;

	while (fgets(buf, sizeof buf, fp))
	{
		char   *val;

		if ((val = getval(buf, "EXEC=")))
		{
			selected = strdup(val);
			break;
		}
	}
	fclose(fp);
}

typedef struct
{
	char   *name;
	char   *icon;
	char   *exec;
	GdkPixmap *pixmap;
	GdkBitmap *mask;
}
wm_entry;

static void
ReadWMConfig(wm_entry *entry, char *fname, GdkWindow *window, GtkStyle *style)
{
	FILE   *fp;
	char    buf[BUFSIZ];

	memset(entry, 0, sizeof(*entry));

	if (!fname)
		return;

	fp = fopen(fname, "r");
	if (!fp)
		return;

	while (fgets(buf, sizeof buf, fp))
	{
		char   *val;

		if ((val = getval(buf, "NAME=")))
		{
			entry->name = strdup(val);
			continue;
		}
		if ((val = getval(buf, "ICON=")))
		{
			if (access(val, R_OK) < 0)
				error(EXIT_SUCCESS, errno, "%s: bad icon %s",
				      fname, val);
			else
				entry->icon = strdup(val);
			continue;
		}
		if ((val = getval(buf, "EXEC=")))
		{
			if (access(val, X_OK) < 0)
				error(EXIT_SUCCESS, errno,
				      "%s: bad executable %s", fname, val);
			else
				entry->exec = strdup(val);
			continue;
		}
	}
	fclose(fp);

	if (entry->icon)
	{
		errno = 0;
		entry->pixmap =
			gdk_pixmap_create_from_xpm(window, &entry->mask,
						   &style->
						   bg[GTK_STATE_NORMAL],
						   entry->icon);
		if (!entry->pixmap)
		{
			error(EXIT_SUCCESS, errno, "Error loading pixmap: %s",
			      entry->icon);
		}
	}
}

static int
CheckEntry(const struct dirent *ent)
{
	struct stat stb;
	char   *path;

	if (strchr(ent->d_name, '.'))
		return 0;

	xasprintf(&path, "%s/%s", WM_DIR, ent->d_name);

	if (lstat(path, &stb) < 0)
		return 0;
	free(path);

	if (!S_ISREG(stb.st_mode))
		return 0;

	return 1;
}

static const char *
getLine(const char *fname, const char *def)
{
	char    buf[BUFSIZ];
	char   *line = 0;
	FILE   *fp = fopen(fname, "r");

	if (!fp)
		return def;

	if (fgets(buf, sizeof buf, fp))
	{
		size_t  len = strlen(buf);

		if (len && ('\n' == buf[len - 1]))
			buf[--len] = '\0';
		if (len)
			line = strdup(buf);
	}

	fclose(fp);
	return line ? : def;
}

static void
select_handler(GtkWidget *widget __attribute__ ((unused)), const char *data)
{
	selected = data;
	gtk_main_quit();
}

static void
pref_handler(GtkWidget *widget __attribute__ ((unused)), GtkWidget * button)
{
	save_prefs = GTK_TOGGLE_BUTTON(button)->active;
}

static FONT_TYPE *
LoadFont(const char *env_name, const char *font_name)
{
	FONT_TYPE *font;
	const char *name = getenv(env_name);

	if (!name || !*name)
		name = font_name;

	font = FONT_FUNC(name);
	if (!font)
		error(EXIT_FAILURE, errno ? : ENOENT, "LoadFont: %s", name);

	return font;
}

static void
SetFont(GtkWidget * widget, const char *env_name, const char *font_name)
{
	GtkStyle *style = gtk_widget_get_style(widget);

	style->FONT_ATTR = LoadFont(env_name, font_name);

	gtk_widget_set_style(widget, style);
}

static GtkWidget *
MakeButton(wm_entry *entry)
{
	GtkWidget *vbox;
	GtkWidget *label;
	GtkWidget *pixmapwid;
	GtkWidget *button;

	vbox = gtk_vbox_new(0, 0);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 1);

	pixmapwid = gtk_pixmap_new(entry->pixmap, entry->mask);
	gtk_box_pack_start(GTK_BOX(vbox), pixmapwid, 0, 0, 5);
	gtk_widget_show(pixmapwid);

	label = gtk_label_new(entry->name);
	gtk_box_pack_end(GTK_BOX(vbox), label, 0, 0, 5);
	gtk_widget_show(label);

	gtk_widget_show(vbox);

	button = gtk_button_new();
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
			   GTK_SIGNAL_FUNC(select_handler),
			   (gpointer) entry->exec);
	gtk_container_add(GTK_CONTAINER(button), vbox);
	gtk_widget_show(button);

	return button;
}

static void
FillBox(GtkWidget * parent, GdkWindow * window, GtkStyle * style)
{
	GtkWidget *hbox = gtk_hbox_new(1, 0);
	GtkWidget *button;
	wm_entry  *entry_list;
	int     i, total = 0, wrap_i = 0, wrap_limit;

	struct dirent **namelist;
	int     n = scandir(WM_DIR, &namelist, CheckEntry, alphasort);

	if (n <= 0)
		error(EXIT_FAILURE, errno, "scandir: %s", WM_DIR);

	if (!(entry_list = calloc((unsigned)n, sizeof(*entry_list))))
		error(EXIT_FAILURE, errno, "calloc");
	for (i = 0; i < n; ++i)
	{
		wm_entry entry;
		char   *fname;

		xasprintf(&fname, "%s/%s", WM_DIR, namelist[i]->d_name);
		free(namelist[i]);
		ReadWMConfig(&entry, fname, window, style);
		if (!entry.name || !entry.icon || !entry.exec || !entry.pixmap)
		{
			error(EXIT_SUCCESS, 0, "%s: incomplete config",
			      fname);
			free(fname);
			free(entry.name);
			free(entry.icon);
			free(entry.exec);
			continue;
		}
		free(fname);

		entry_list[total++] = entry;

		if (!preferred)
			preferred = entry.exec;
	}
	free(namelist);

	if (total <= 6)
		wrap_limit = total;
	else
	{
		if (total & 1)
			wrap_limit = (total + 1) / 2;
		else
			wrap_limit = total / 2;
	}
	// error (EXIT_SUCCESS, 0, "total=%d, wrap_limit=%d", total, wrap_limit);
	for (i = 0; i < total; ++i)
	{
		wm_entry *entry = &entry_list[i];

		button = MakeButton(entry);
		free(entry->icon);
		gtk_box_pack_start(GTK_BOX(hbox), button, 0, 1, 0);
		gtk_widget_show(button);
		if (++wrap_i >= wrap_limit)
		{
			gtk_box_pack_start(GTK_BOX(parent), hbox, 0, 0, 0);
			gtk_widget_show(hbox);
			hbox = gtk_hbox_new(1, 0);
			wrap_i = 0;
		}
	}
	free(entry_list);
	if (wrap_i)
	{
		gtk_box_pack_start(GTK_BOX(parent), hbox, 0, 0, 0);
		gtk_widget_show(hbox);
	}

}

static GtkWidget *
MakeWelcomeLabel(GtkWidget * vbox, const char *text)
{
	GtkWidget *label = gtk_label_new(text);

	SetFont(label, "WM_SELECT_WELCOME_FONT", WM_SELECT_WELCOME_FONT);
	gtk_misc_set_alignment(GTK_MISC(label), 0.5, 0.5);

	gtk_box_pack_start(GTK_BOX(vbox), label, 0, 0, 10);
	gtk_widget_show(label);

	return label;
}

static void
MakeWindows()
{
	GtkWidget *window;
	GtkWidget *button;
	GtkWidget *vbox;
	GtkWidget *hbox;
	GtkWidget *separator;

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER_ALWAYS);
	gtk_container_set_border_width(GTK_CONTAINER(window), 10);
	gtk_window_set_title(GTK_WINDOW(window), "WM Selector");

	gtk_signal_connect(GTK_OBJECT(window), "destroy",
			   GTK_SIGNAL_FUNC(gtk_main_quit),
			   GTK_OBJECT(window));
	gtk_signal_connect(GTK_OBJECT(window), "delete_event",
			   GTK_SIGNAL_FUNC(gtk_main_quit),
			   GTK_OBJECT(window));

	gtk_widget_realize(window);

	SetFont(window, "WM_SELECT_LABEL_FONT", WM_SELECT_LABEL_FONT);

	vbox = gtk_vbox_new(0, 0);

	MakeWelcomeLabel(vbox, _("Welcome to"));

	FillBox(vbox, window->window, gtk_widget_get_style(window));

	MakeWelcomeLabel(vbox, getLine("/etc/altlinux-release", "ALT Linux"));

	separator = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(vbox), separator, 0, 0, 10);
	gtk_widget_show(separator);

	hbox = gtk_hbox_new(0, 0);

	button = gtk_check_button_new_with_label(_
						 ("Set default window manager"));
	gtk_signal_connect_object(GTK_OBJECT(button), "toggled",
				  GTK_SIGNAL_FUNC(pref_handler),
				  GTK_OBJECT(button));
	gtk_box_pack_start(GTK_BOX(hbox), button, 1, 1, 0);
	gtk_widget_show(button);

	button = gtk_button_new_with_label(_("Exit"));
	GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
	gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
				  GTK_SIGNAL_FUNC(gtk_main_quit),
				  GTK_OBJECT(window));
	gtk_box_pack_end(GTK_BOX(hbox), button, 0, 0, 10);
	gtk_widget_show(button);

	gtk_box_pack_end(GTK_BOX(vbox), hbox, 0, 0, 0);
	gtk_widget_show(hbox);

	gtk_container_add(GTK_CONTAINER(window), vbox);
	gtk_widget_show(vbox);

	gtk_widget_show(window);
	gtk_widget_grab_default(button);
}

static void
InitLocale()
{
	const char *name = getenv("NLSPATH");

	if (!name || !*name)
		name = LOCALEDIR;

	if (!gtk_set_locale())
		error(EXIT_FAILURE, errno, "gtk_set_locale");

	if (!textdomain(PACKAGE))
		error(EXIT_FAILURE, errno, "textdomain");

	if (!bindtextdomain(PACKAGE, name))
		error(EXIT_FAILURE, errno, "bindtextdomain");

#ifdef WITH_GTK2
	if (!bind_textdomain_codeset(PACKAGE, "UTF-8"))
		error(EXIT_FAILURE, errno, "bind_textdomain_codeset");
#endif /* WITH_GTK2 */
}

int
main(int argc, char *argv[])
{
	InitLocale();

	ReadProgramConfig();
	ExecSelected();

	gtk_init(&argc, &argv);

	MakeWindows();
	gtk_main();

	ExecSelected();
	return EXIT_FAILURE;
}
